#!/usr/bin/env python
##
# Copyright (c) 2000 by g.p.ciceri <gp.ciceri@acm.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# Alternatively, you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#                                   
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU (Library) General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.                          
#
#
# $Author: zanzi $
# $Revision: 1.1 $
# $Log: BasicOptions.py,v $
# Revision 1.1  2000/01/02 17:42:34  zanzi
# Initial revision
#
#
# module name: BasicOptions.py
#
# description: python functions to handle basic options prices algorithms
#
#

from types import ListType, TupleType
from math import exp, sqrt, log
from NormDist import n, N, NRho

def OptionPriceCallBlackScholes(spotPrice, strikePrice, interestRate, sigma, time):
    timeSqrt = sqrt(time)
    d1 = (log(spotPrice/strikePrice) + (interestRate * time)) / ( sigma * timeSqrt ) + (0.5 * sigma * timeSqrt)
    d2 = d1 - (sigma * timeSqrt)
    c = spotPrice * N(d1) - strikePrice * exp(-interestRate * time) * N(d2)
    return c

def OptionPricePutBlackScholes(spotPrice, strikePrice, interestRate, sigma, time):
    timeSqrt = sqrt(time)
    d1 = (log(spotPrice/strikePrice) + (interestRate * time)) / ( sigma * timeSqrt ) + (0.5 * sigma * timeSqrt)
    d2 = d1 - (sigma * timeSqrt)
    p = - spotPrice * N(-d1) + strikePrice * exp(-interestRate * time) * N(-d2)
    return p

def OptionPriceDeltaCallBlackScholes(spotPrice, strikePrice, interestRate, sigma, time):
    timeSqrt = sqrt(time)
    d1 = (log(spotPrice/strikePrice) + (interestRate * time)) / ( sigma * timeSqrt ) + (0.5 * sigma * timeSqrt)
    delta = N(d1)
    return delta

def OptionPriceDeltaPutBlackScholes(spotPrice, strikePrice, interestRate, sigma, time):
    timeSqrt = sqrt(time)
    d1 = (log(spotPrice/strikePrice) + (interestRate * time)) / ( sigma * timeSqrt ) + (0.5 * sigma * timeSqrt)
    delta = -N(-d1)
    return delta

def OptionPricePartialsCallBlackScholes(spotPrice, strikePrice, interestRate, sigma, time):
    timeSqrt = sqrt(time)
    d1 = (log(spotPrice/strikePrice) + (interestRate * time)) / ( sigma * timeSqrt ) + (0.5 * sigma * timeSqrt)
    d2 = d1 - (sigma * timeSqrt)
    delta = N(d1)
    gamma = n(d1)/(spotPrice * sigma * timeSqrt)
    theta = - (spotPrice * sigma * n(d1)) / (2 * timeSqrt) - interestRate * strikePrice * exp( - interestRate * time ) * N(d2)
    vega  = spotPrice * timeSqrt * n(d1)
    rho   = strikePrice * time * exp( - interestRate * time ) * N(d2)
    return (delta, gamma, theta, vega, rho)

def OptionPriceImpliedVolatilityCallBlackScholesBisections(spotPrice, strikePrice, interestRate, time, optionPrice):
    sigmaLow = 0.0001
    price = OptionPriceCallBlackScholes(spotPrice, strikePrice, interestRate, sigmaLow, time)
    if (price > optionPrice): return 0.0

    ACCURACY       = 1.0e-05
    MAX_ITERATIONS = 100
    HIGH_VALUE     = 1e10
    ERROR          = -1e40

    sigmaHigh = 0.3
    price = OptionPriceCallBlackScholes(spotPrice, strikePrice, interestRate, sigmaHigh, time)
    while (price < optionPrice):
        sigmaHigh = 2.0 * sigmaHigh
        price = OptionPriceCallBlackScholes(spotPrice, strikePrice, interestRate, sigmaHigh, time)
        if (sigmaHigh > HIGH_VALUE): return ERROR
    for x in xrange(MAX_ITERATIONS):
        sigma = (sigmaLow + sigmaHigh) * 0.5
        price = OptionPriceCallBlackScholes(spotPrice, strikePrice, interestRate, sigma, time)
        test = price - optionPrice
        if (abs(test) < ACCURACY): return sigma
        if (test < 0.0): sigmaLow = sigma
        else: sigmaHigh = sigma
    return ERROR
        
def OptionPriceImpliedVolatilityCallBlackScholesNewton(spotPrice, strikePrice, interestRate, time, optionPrice):
    sigmaLow = 1e-5
    price = OptionPriceCallBlackScholes(spotPrice, strikePrice, interestRate, sigmaLow, time)
    if (price > optionPrice): return 0.0

    ACCURACY       = 1.0e-04
    MAX_ITERATIONS = 100
    ERROR          = -1.0e40
    
    timeSqrt = sqrt(time)
    sigma = (optionPrice / spotPrice) / (0.398 * timeSqrt)
    for x in xrange(MAX_ITERATIONS):
        price = OptionPriceCallBlackScholes(spotPrice, strikePrice, interestRate, sigma, time)
        diff = price - optionPrice
        if (abs(diff) < ACCURACY): return sigma
        d1 = (log(spotPrice/strikePrice) + (interestRate * time)) / ((sigma * timeSqrt) + (0.5 * sigma * timeSqrt))
        vega = spotPrice * timeSqrt * n(d1)
        sigma = sigma + diff/vega
    return ERROR

def OptionPriceEuropeanCallPayout(spotPrice, strikePrice, interestRate, yield, sigma, time):
    sigmaSqr = pow(sigma,2)
    timeSqrt = sqrt(time)
    d1 = (log(spotPrice/strikePrice) + ((interestRate-yield + 0.5 * sigmaSqr) * time)) / (sigma * timeSqrt)
    d2 = d1 - (sigma * timeSqrt)
    callPrice = spotPrice * exp(-yield*time) * N(d1) - strikePrice * exp(-interestRate*time) * N(d2)
    return callPrice

def OptionPriceEuropeanPutPayout(spotPrice, strikePrice, interestRate, yield, sigma, time):
    sigmaSqr = pow(sigma,2)
    timeSqrt = sqrt(time)
    d1 = (log(spotPrice/strikePrice) + ((interestRate-yield + 0.5 * sigmaSqr) * time)) / (sigma * timeSqrt)
    d2 = d1 - (sigma * timeSqrt)
    putPrice =  - spotPrice * exp(-yield*time) * N(-d1) + strikePrice * exp(-interestRate*time) * N(-d2)
    return putPrice

def OptionPriceEuropeanCallDividends(spotPrice, strikePrice, interestRate, sigma, time, dividendTimes, dividendAmounts):
    for theParameter in (dividendTimes, dividendAmounts):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
    if not len(dividendTimes) == len(dividendAmounts):
        raise StandardError, "times and amount must be of the same length"
    
    for x in xrange(len(dividendTimes)):
        if (dividendTimes[x] < time):
            spotPrice = spotPrice - dividendAmounts[x] * exp(-interestRate * dividendTimes[x])
 
    return OptionPriceCallBlackScholes(spotPrice, strikePrice, interestRate, sigma, time)

def OptionPriceEuropeanPutDividends(spotPrice, strikePrice, interestRate, sigma, time, dividendTimes, dividendAmounts):
    for theParameter in (dividendTimes, dividendAmounts):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
    if not len(dividendTimes) == len(dividendAmounts):
        raise StandardError, "times and amount must be of the same length"
    
    for x in xrange(len(dividendTimes)):
        if (dividendTimes[x] < time):
            spotPrice = spotPrice - dividendAmounts[x] * exp(-interestRate * dividendTimes[x])
 
    return OptionPricePutBlackScholes(spotPrice, strikePrice, interestRate, sigma, time)

def OptionPriceAmericanCallDividend(spotPrice, strikePrice, interestRate, sigma, tau, d1, tau1):
    
    if (d1 <= strikePrice * (1.0 - exp(-interestRate * (tau - tau1)))):
        newSpotPrice = spotPrice - exp(-interestRate * tau1) * d1
        return OptionPriceCallBlackScholes(newSpotPrice, strikePrice, interestRate, sigma, tau)
    
    ACCURACY = 1e-06
    sigmaSqr = sigma*sigma
    tauSqrt  = sqrt(tau)
    tau1Sqrt = sqrt(tau1)
    rho      = - sqrt (tau1/tau)

    spotPriceBar  = 0
    spotPriceLow  = 0
    spotPriceHigh = spotPrice
    c = OptionPriceCallBlackScholes(spotPriceHigh, strikePrice, interestRate,sigma, (tau - tau1))
    test = c - spotPriceHigh - d1 + strikePrice

    while ( (test > 0.0) and (spotPriceHigh <= 1e10) ):
        spotPriceHigh = spotPriceHigh * 2.0
        c = OptionPriceCallBlackScholes(spotPriceHigh, strikePrice, interstRate,sigma, (tau - tau1))
        test = c - spotPriceHigh - d1 + strikePrice

    if (spotPriceHigh > 1e10):
        newSpotPrice = spotPrice - exp(-interestRate * tau1) * d1
        return OptionPriceCallBlackScholes(newSpotPrice, strikePrice, interestRate, sigma, tau)

    spotPriceBar = 0.5 * spotPriceHigh
    c = OptionPriceCallBlackScholes(spotPriceBar, strikePrice, interestRate,sigma, (tau - tau1))
    test = c - spotPriceBar - d1 + strikePrice

    while ( (abs(test) > ACCURACY) and ((spotPriceHigh-spotPriceLow)) > ACCURACY):
        if (test < 0.0): spotPriceHigh = spotPriceBar
        else: spotPriceLow = spotPriceBar

        spotPriceBar = 0.5 * (spotPriceHigh + spotPriceLow)
        c = OptionPriceCallBlackScholes(spotPriceBar, strikePrice, interestRate,sigma, (tau - tau1))
        test = c - spotPriceBar - d1 + strikePrice
    a1 = (log((spotPrice - d1 * exp(-interestRate * tau1))/strikePrice) + (interestRate + 0.5*sigmaSqr) * tau) / (sigma * tauSqrt)
    a2 = a1 - sigma * tauSqrt
    b1 = (log((spotPrice - d1 * exp(-interestRate * tau1))/spotPriceBar) + (interestRate + 0.5*sigmaSqr) * tau1) / (sigma * tau1Sqrt)
    b2 = b1 - sigma * tau1Sqrt
    c  = (spotPrice - d1 * exp(-interestRate*tau1)) * N(b1)                \
         + (spotPrice - d1 * exp(-interestRate*tau1)) * NRho(a1, -b1, rho) \
         - (strikePrice * exp(-interestRate*tau)) * NRho(a2, -b2, rho)    \
         - (strikePrice - d1) * exp(-interestRate*tau1) * N(b2)
    return c


if __name__ == '__main__':
    spot = 58.507
    exercise = 60
    r = 0.06
    sigma = 0.35
    timeToMaturity = 0.25
    print 'call price: ', OptionPriceCallBlackScholes(spot, exercise, r, sigma, timeToMaturity)
    print 'put price: ', OptionPricePutBlackScholes(spot, exercise, r, sigma, timeToMaturity)
    print 'call delta: ', OptionPriceDeltaCallBlackScholes(spot, exercise, r, sigma, timeToMaturity)
    print 'put delta: ', OptionPriceDeltaPutBlackScholes(spot, exercise, r, sigma, timeToMaturity)
    print 'continous payout 5% '
    spot = 100.0
    exercise = 100.0
    r = 0.1
    sigma = 0.25
    timeToMaturity = 1.0
    yield = 0.05
    print 'call price: ', OptionPriceEuropeanCallPayout(spot, exercise, r, yield, sigma, timeToMaturity)
    print 'put price: ', OptionPriceEuropeanPutPayout(spot, exercise, r, yield, sigma, timeToMaturity)
    print 'discrete dividends, european options'
    dividendTimes = [0.5]
    dividendAmounts = [2.0]
    print 'call price', OptionPriceEuropeanCallDividends(spot, exercise, r, sigma, timeToMaturity, dividendTimes, dividendAmounts)
    print 'put price', OptionPriceEuropeanPutDividends(spot, exercise, r, sigma, timeToMaturity, dividendTimes, dividendAmounts)
    print 'American call, one dividend, exact solution '
    D=2
    timeToDividend=0.5
    print 'call price (D=2)', OptionPriceAmericanCallDividend(spot, exercise, r, sigma, timeToMaturity, D, timeToDividend)
    D=10
    print 'call price (D=10)', OptionPriceAmericanCallDividend(spot, exercise, r, sigma, timeToMaturity, D, timeToDividend)






