#!/usr/bin/env python
##
# Copyright (c) 2000 by g.p.ciceri <gp.ciceri@acm.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# Alternatively, you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#                                   
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU (Library) General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.                    
#
#
# $Author: zanzi $
# $Revision: 1.1 $
# $Log: BinomialOptions.py,v $
# Revision 1.1  2000/01/02 17:42:52  zanzi
# Initial revision
#
#
# module name: BinomialOptions.py
#
# description: python functions to handle binomial options prices algorithms
#
#

from types import ListType, TupleType
from math import exp, sqrt, log
from NormDist import n, N, NRho

def OptionPriceCallEuropeanBinomial(spotPrice, strikePrice, interestRate, sigma, time, stepsNum):
    R = exp(interestRate * (time / stepsNum))
    Rinv = 1.0 / R
    u = exp(sigma * sqrt(time/stepsNum))
    uu = u * u
    d = 1.0 / u
    pUp = (R - d)/(u - d)
    pDown = 1.0 - pUp
    prices = []
    prices.append(spotPrice * pow(d, stepsNum))
    for i in xrange(1, stepsNum+1):
        prices.append(uu * prices[i-1])
    callValues = []
    for i in xrange(stepsNum+1):
        callValues.append(max(0.0, (prices[i] - strikePrice)))
    for i in xrange(stepsNum, 0, -1):
        for j in xrange(i):
            callValues[j] = (pUp * callValues[j+1] + pDown * callValues[j]) * Rinv
    return callValues[0]

def OptionPricePutEuropeanBinomial(spotPrice, strikePrice, interestRate, sigma, time, stepsNum):
    R = exp(interestRate * (time / stepsNum))
    Rinv = 1.0 / R
    u = exp(sigma * sqrt(time/stepsNum))
    uu = u * u
    d = 1.0 / u
    pUp = (R - d)/(u - d)
    pDown = 1.0 - pUp
    prices = []
    prices.append(spotPrice * pow(d, stepsNum))
    for i in xrange(1, stepsNum+1):
        prices.append(uu * prices[i-1])
    putValues = []
    for i in xrange(stepsNum+1):
        putValues.append(max(0.0, ( - prices[i] + strikePrice)))
    for i in xrange(stepsNum, 0, -1):
        for j in xrange(i):
            putValues[j] = (pUp * putValues[j+1] + pDown * putValues[j]) * Rinv
    return putValues[0]

def OptionPriceCallAmericanBinomial(spotPrice, strikePrice, interestRate, sigma, time, stepsNum):
    R = exp(interestRate * (time / stepsNum))
    Rinv = 1.0 / R
    u = exp(sigma * sqrt(time/stepsNum))
    uu = u * u
    d = 1.0 / u
    pUp = (R - d)/(u - d)
    pDown = 1.0 - pUp
    prices = []
    callValues = []
    prices.append(spotPrice * pow(d, stepsNum))
    for i in xrange(1, stepsNum+1):
        prices.append(uu * prices[i-1])
    for i in xrange(stepsNum+1):
        callValues.append(max(0.0, (prices[i] - strikePrice)))
    for i in xrange(stepsNum, 0, -1):
        for j in xrange(i):
            callValues[j] = (pUp * callValues[j+1] + pDown * callValues[j]) * Rinv
            prices[j] = d * prices[j+1]
            callValues[j] = max(callValues[j], (prices[j] - strikePrice))
    return callValues[0]

def OptionPricePutAmericanBinomial(spotPrice, strikePrice, interestRate, sigma, time, stepsNum):
    R = exp(interestRate * (time / stepsNum))
    Rinv = 1.0 / R
    u = exp(sigma * sqrt(time/stepsNum))
    uu = u * u
    d = 1.0 / u
    pUp = (R - d)/(u - d)
    pDown = 1.0 - pUp
    prices = []
    putValues = []
    prices.append(spotPrice * pow(d, stepsNum))
    for i in xrange(1, stepsNum+1):
        prices.append(uu * prices[i-1])
    for i in xrange(stepsNum+1):
        putValues.append(max(0.0, (- prices[i] + strikePrice)))
    for i in xrange(stepsNum, 0, -1):
        for j in xrange(i):
            putValues[j] = (pUp * putValues[j+1] + pDown * putValues[j]) * Rinv
            prices[j] = d * prices[j+1]
            putValues[j] = max(putValues[j], (- prices[j] + strikePrice))
    return putValues[0]

def OptionPriceDeltaAmericanCallBinomial(spotPrice, strikePrice, interestRate, sigma, time, stepsNum):
    prices = []
    callValues = []   
    R = exp(interestRate * (time / stepsNum))
    Rinv = 1.0 / R
    u = exp(sigma * sqrt(time/stepsNum))
    uu = u * u
    d = 1.0 / u
    pUp = (R - d)/(u - d)
    pDown = 1.0 - pUp
    prices.append(spotPrice * pow(d, stepsNum))
    for i in xrange(1, stepsNum+1):
        prices.append(uu * prices[i-1])
    for i in xrange(stepsNum+1):
        callValues.append(max(0.0, (prices[i] - strikePrice)))
    for i in xrange(stepsNum, 1, -1):
        for j in xrange(i):
            prices[j] = d * prices[j+1]
            callValues[j] = (pDown * callValues[j] + pUp * callValues[j+1]) * Rinv
            callValues[j] = max(callValues[j], (prices[j] - strikePrice))
    delta = (callValues[1] - callValues[0])/(spotPrice * u - spotPrice * d)
    return delta

def OptionPriceDeltaAmericanPutBinomial(spotPrice, strikePrice, interestRate, sigma, time, stepsNum):
    prices = []
    putValues = []   
    R = exp(interestRate * (time / stepsNum))
    Rinv = 1.0 / R
    u = exp(sigma * sqrt(time/stepsNum))
    uu = u * u
    d = 1.0 / u
    pUp = (R - d)/(u - d)
    pDown = 1.0 - pUp
    prices.append(spotPrice * pow(d, stepsNum))
    for i in xrange(1, stepsNum+1):
        prices.append(uu * prices[i-1])
    for i in xrange(stepsNum+1):
        putValues.append(max(0.0, (- prices[i] + strikePrice)))
    for i in xrange(stepsNum, 1, -1):
        for j in xrange(i):
            prices[j] = d * prices[j+1]
            putValues[j] = (pDown * putValues[j] + pUp * putValues[j+1]) * Rinv
            putValues[j] = max(putValues[j], ( - prices[j] + strikePrice))
    delta = (putValues[1] - putValues[0])/(spotPrice * u - spotPrice * d)
    return delta

def OptionPricePartialsAmericanCallBinomial(spotPrice, strikePrice, interestRate, sigma, time, stepsNum):
    prices = []
    callValues = []   
    R = exp(interestRate * (time/stepsNum))
    Rinv = 1.0 / R
    u = exp(sigma * sqrt(time/stepsNum))
    uu = u * u
    d = 1.0 / u
    pUp = (R - d)/(u - d)
    pDown = 1.0 - pUp
    prices.append(spotPrice * pow(d, stepsNum))
    
    for i in xrange(1, stepsNum+1):
        prices.append(uu * prices[i-1])
    for i in xrange(stepsNum+1):
        callValues.append(max(0.0, (prices[i] - strikePrice)))
    for i in xrange(stepsNum, 2, -1):
        for j in xrange(i):
            prices[j] = d * prices[j+1]
            callValues[j] = (pDown * callValues[j] + pUp * callValues[j+1]) * Rinv
            callValues[j] = max(callValues[j], (prices[j] - strikePrice))

    f22 = callValues[2]
    f21 = callValues[1]
    f20 = callValues[0]

    prices[0] = d * prices[1]
    callValues[0] = (pDown * callValues[0] + pUp * callValues[1]) * Rinv
    callValues[0] = max(callValues[0], (prices[0] - strikePrice))
    prices[1] = d * prices[2]
    callValues[1] = (pDown * callValues[1] + pUp * callValues[2]) * Rinv
    callValues[1] = max(callValues[1], (prices[1] - strikePrice))

    f11 = callValues[1]
    f10 = callValues[0]

    prices[0] = d * prices[1]
    callValues[0] = (pDown * callValues[0] + pUp * callValues[1]) * Rinv
    callValues[0] = max(callValues[0], (spotPrice - strikePrice))

    f00 = callValues[0]

    delta = (f11 - f10) / (spotPrice * u - spotPrice * d)
    h = 0.5 * spotPrice * (uu - d*d)
    gamma = ((f22 - f21) / (spotPrice * (uu-1)) - (f21 - f20)/(spotPrice * (1 - d*d)))/h
    theta = (f21 - f00) / (2 * (time/stepsNum))

    diff = 0.02
    sigmaTemp = sigma + diff
    pricesTemp = OptionPriceCallAmericanBinomial(spotPrice, strikePrice, interestRate, sigmaTemp, time, stepsNum)
    vega = (pricesTemp - f00) / diff
    diff = 0.05
    rTemp = r + diff
    pricesTemp = OptionPriceCallAmericanBinomial(spotPrice, strikePrice, rTemp, sigma, time, stepsNum)
    rho = (pricesTemp - f00) / diff
    return (delta, gamma, theta, vega, rho)

def OptionPricePartialsAmericanPutBinomial(spotPrice, strikePrice, interestRate, sigma, time, stepsNum):
    prices = []
    putValues = []   
    R = exp(interestRate * (time/stepsNum))
    Rinv = 1.0 / R
    u = exp(sigma * sqrt(time/stepsNum))
    uu = u * u
    d = 1.0 / u
    pUp = (R - d)/(u - d)
    pDown = 1.0 - pUp
    prices.append(spotPrice * pow(d, stepsNum))
    
    for i in xrange(1, stepsNum+1):
        prices.append(uu * prices[i-1])
    for i in xrange(stepsNum+1):
        putValues.append(max(0.0, ( - prices[i] + strikePrice)))
    for i in xrange(stepsNum, 2, -1):
        for j in xrange(i):
            prices[j] = d * prices[j+1]
            putValues[j] = (pDown * putValues[j] + pUp * putValues[j+1]) * Rinv
            putValues[j] = max(putValues[j], ( - prices[j] + strikePrice))

    f22 = putValues[2]
    f21 = putValues[1]
    f20 = putValues[0]

    prices[0] = d * prices[1]
    putValues[0] = (pDown * putValues[0] + pUp * putValues[1]) * Rinv
    putValues[0] = max(putValues[0], ( - prices[0] + strikePrice))
    prices[1] = d * prices[2]
    putValues[1] = (pDown * putValues[1] + pUp * putValues[2]) * Rinv
    putValues[1] = max(putValues[1], (- prices[1] + strikePrice))

    f11 = putValues[1]
    f10 = putValues[0]

    prices[0] = d * prices[1]
    putValues[0] = (pDown * putValues[0] + pUp * putValues[1]) * Rinv
    putValues[0] = max(putValues[0], (spotPrice - strikePrice))

    f00 = putValues[0]

    delta = (f11 - f10) / (spotPrice * u - spotPrice * d)
    h = 0.5 * spotPrice * (uu - d*d)
    gamma = ((f22 - f21) / (spotPrice * (uu-1)) - (f21 - f20)/(spotPrice * (1 - d*d)))/h
    theta = (f21 - f00) / (2 * (time/stepsNum))

    diff = 0.02
    sigmaTemp = sigma + diff
    pricesTemp = OptionPricePutAmericanBinomial(spotPrice, strikePrice, interestRate, sigmaTemp, time, stepsNum)
    vega = (pricesTemp - f00) / diff
    diff = 0.05
    rTemp = r + diff
    pricesTemp = OptionPricePutAmericanBinomial(spotPrice, strikePrice, rTemp, sigma, time, stepsNum)
    rho = (pricesTemp - f00) / diff
    return (delta, gamma, theta, vega, rho)

def OptionPriceCallAmericanProportionalDividendsBinomial(spotPrice, strikePrice, interestRate, sigma, time, stepsNum, dividendTimes, dividendYields):
    for theParameter in (dividendTimes, dividendYields):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
    if not len(dividendTimes) == len(dividendYields):
        raise StandardError, "times and yields must be of the same length"

    if (len(dividendTimes) == 0):
        return OptionPriceCallAmericanBinomial(spotPrice, strikePrice, interestRate, sigma, time, stepsNum)
    
    R = exp(interestRate * (time / stepsNum))
    Rinv = 1.0 / R
    u = exp(sigma * sqrt(time/stepsNum))
    uu = u * u
    d = 1.0 / u
    pUp = (R - d)/(u - d)
    pDown = 1.0 - pUp
    dividendSteps = []
    for i in xrange(len(dividendTimes)):
        dividendSteps.append(int(dividendTimes[i]/time*stepsNum))
    prices = []
    callPrices = []
    prices.append(spotPrice * pow(d, stepsNum))
    for i in xrange(len(dividendTimes)):
        prices[0] = prices[0] * (1.0 - dividendYields[i])
    for i in xrange(1, stepsNum+1):
        prices.append(uu * prices[i-1])
    for i in xrange(stepsNum+1):
        callPrices.append(max(0.0, (prices[i] - strikePrice)))
    for step in xrange(stepsNum-1, -1, -1):
        for i in xrange(len(dividendTimes)):
            if (step == dividendSteps[i]):
                for j in xrange(step):
                    prices[j] = prices[j] * (1.0/(1.0 - dividendYields[i]))
        for i in xrange(step+1):
            prices[i] = d * prices[i+1]
            callPrices[i] = (pDown * callPrices[i] + pUp * callPrices[i+1]) * Rinv
            callPrices[i] = max(callPrices[i], prices[i] - strikePrice)
    return callPrices[0]

def OptionPricePutAmericanProportionalDividendsBinomial(spotPrice, strikePrice, interestRate, sigma, time, stepsNum, dividendTimes, dividendYields):
    for theParameter in (dividendTimes, dividendYields):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
    if not len(dividendTimes) == len(dividendYields):
        raise StandardError, "times and yields must be of the same length"

    if (len(dividendTimes) == 0):
        return OptionPriceCallAmericanBinomial(spotPrice, strikePrice, interestRate, sigma, time, stepsNum)
    
    R = exp(interestRate * (time / stepsNum))
    Rinv = 1.0 / R
    u = exp(sigma * sqrt(time/stepsNum))
    uu = u * u
    d = 1.0 / u
    pUp = (R - d)/(u - d)
    pDown = 1.0 - pUp
    dividendSteps = []
    for i in xrange(len(dividendTimes)):
        dividendSteps.append(int(dividendTimes[i]/time*stepsNum))
    prices = []
    putPrices = []
    prices.append(spotPrice * pow(d, stepsNum))
    for i in xrange(len(dividendTimes)):
        prices[0] = prices[0] * (1.0 - dividendYields[i])
    for i in xrange(1, stepsNum+1):
        prices.append(uu * prices[i-1])
    for i in xrange(stepsNum+1):
        putPrices.append(max(0.0, ( - prices[i] + strikePrice)))
    for step in xrange(stepsNum-1, -1, -1):
        for i in xrange(len(dividendTimes)):
            if (step == dividendSteps[i]):
                for j in xrange(step):
                    prices[j] = prices[j] * (1.0/(1.0 - dividendYields[i]))
        for i in xrange(step+1):
            prices[i] = d * prices[i+1]
            putPrices[i] = (pDown * putPrices[i] + pUp * putPrices[i+1]) * Rinv
            putPrices[i] = max(putPrices[i],  - prices[i] + strikePrice)
    return putPrices[0]

def OptionPriceCallAmericanDiscreteDividendsBinomial(spotPrice, strikePrice, interestRate, sigma, time, stepsNum, dividendTimes, dividendAmounts):
    for theParameter in (dividendTimes, dividendAmounts):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
    if not len(dividendTimes) == len(dividendAmounts):
        raise StandardError, "times and amounts must be of the same length"

    if (len(dividendTimes) == 0):
        return OptionPriceCallAmericanBinomial(spotPrice, strikePrice, interestRate, sigma, time, stepsNum)

    stepsBeforeDividend = int(dividendTimes[0]/time*stepsNum)
    R = exp(interestRate * (time / stepsNum))
    Rinv = 1.0 / R
    u = exp(sigma * sqrt(time/stepsNum))
    uu = u * u
    d = 1.0 / u
    pUp = (R - d)/(u - d)
    pDown = 1.0 - pUp
    dividendAmountZero = dividendAmounts[0]

    tmpDividendTimes   = []
    tmpDividendAmounts = []
    prices = []
    callValues = []

    for i in xrange(len(dividendTimes)-1):
        tmpDividendAmounts.append(dividendAmounts[i+1])
        tmpDividendTimes.append(dividendTimes[i+1] - dividendTimes[0])
    prices.append(spotPrice * pow(d, stepsBeforeDividend))
    for i in xrange(1, stepsBeforeDividend+1):
        prices.append(uu * prices[i-1])
    for i in xrange(stepsBeforeDividend+1):
        valueAlive = OptionPriceCallAmericanDiscreteDividendsBinomial(prices[i]-dividendAmountZero, strikePrice, interestRate, sigma, (time - dividendTimes[0]) , (stepsNum - stepsBeforeDividend), tmpDividendTimes, tmpDividendAmounts)
        callValues.append(max(valueAlive, prices[i] - strikePrice))
    for step in xrange(stepsBeforeDividend, -1, -1):
        for i in xrange(step):
            prices[i] = d * prices[i+1]
            callValues[i] = (pDown * callValues[i] + pUp * callValues[i+1]) * Rinv
            callValues[i] = max(callValues[i], prices[i] - strikePrice)
    return callValues[0]

def OptionPricePutAmericanDiscreteDividendsBinomial(spotPrice, strikePrice, interestRate, sigma, time, stepsNum, dividendTimes, dividendAmounts):
    for theParameter in (dividendTimes, dividendAmounts):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
    if not len(dividendTimes) == len(dividendAmounts):
        raise StandardError, "times and amounts must be of the same length"

    if (len(dividendTimes) == 0):
        return OptionPricePutAmericanBinomial(spotPrice, strikePrice, interestRate, sigma, time, stepsNum)

    stepsBeforeDividend = int(dividendTimes[0]/time*stepsNum)
    R = exp(interestRate * (time / stepsNum))
    Rinv = 1.0 / R
    u = exp(sigma * sqrt(time/stepsNum))
    uu = u * u
    d = 1.0 / u
    pUp = (R - d)/(u - d)
    pDown = 1.0 - pUp
    dividendAmountZero = dividendAmounts[0]

    tmpDividendTimes   = []
    tmpDividendAmounts = []
    prices = []
    callValues = []

    for i in xrange(len(dividendTimes)-1):
        tmpDividendAmounts.append(dividendAmounts[i+1])
        tmpDividendTimes.append(dividendTimes[i+1] - dividendTimes[0])
    prices.append(spotPrice * pow(d, stepsBeforeDividend))
    for i in xrange(1, stepsBeforeDividend+1):
        prices.append(uu * prices[i-1])
    for i in xrange(stepsBeforeDividend+1):
        valueAlive = OptionPricePutAmericanDiscreteDividendsBinomial(prices[i]-dividendAmountZero, strikePrice, interestRate, sigma, (time - dividendTimes[0]) , (stepsNum - stepsBeforeDividend), tmpDividendTimes, tmpDividendAmounts)
        callValues.append(max(valueAlive, - prices[i] + strikePrice))
    for step in xrange(stepsBeforeDividend, -1, -1):
        for i in xrange(step):
            prices[i] = d * prices[i+1]
            callValues[i] = (pDown * callValues[i] + pUp * callValues[i+1]) * Rinv
            callValues[i] = max(callValues[i], - prices[i] + strikePrice)
    return callValues[0]

if __name__ == '__main__':
    spot = 100.0
    exercise = 100.0
    r = 0.1
    sigma = 0.25
    timeToMaturity = 1.0
    steps = 100
    print 'european'
    print 'call: ', OptionPriceCallEuropeanBinomial(spot, exercise, r, sigma, timeToMaturity, steps)
    print 'put: ', OptionPricePutEuropeanBinomial(spot, exercise, r, sigma, timeToMaturity, steps)
    
    print 'american'
    print 'call: ', OptionPriceCallAmericanBinomial(spot, exercise, r, sigma, timeToMaturity, steps)
    print 'put: ', OptionPricePutAmericanBinomial(spot, exercise, r, sigma, timeToMaturity, steps)

    print 'Proportional dividends'
    dividendTimes = [0.25, 0.75]
    dividendYields = [0.03, 0.03]
    print 'american call, dividends=3%, 3%, price:', OptionPriceCallAmericanProportionalDividendsBinomial(spot, exercise, r, sigma, timeToMaturity, steps, dividendTimes, dividendYields)
    print 'american put, dividends=3%, 3%, price:', OptionPricePutAmericanProportionalDividendsBinomial(spot, exercise, r, sigma, timeToMaturity, steps, dividendTimes, dividendYields)
    print 'Discrete dividends'
    dividendAmounts = [2.5, 2.5]
    print 'american call, dividends=2.5, 2.5, price:', OptionPriceCallAmericanDiscreteDividendsBinomial(spot, exercise, r, sigma, timeToMaturity, steps, dividendTimes, dividendAmounts)
    print 'american put, dividends=2.5, 2.5, price:', OptionPricePutAmericanDiscreteDividendsBinomial(spot, exercise, r, sigma, timeToMaturity, steps, dividendTimes, dividendAmounts)

    spot = 100.0
    exercise = 100.0
    r = 0.1
    sigma = 0.25
    timeToMaturity = 1.0
    steps = 100
    print 'american call, delta: ', OptionPriceDeltaAmericanCallBinomial(spot, exercise, r, sigma, timeToMaturity, steps)
    print 'american call, put: ', OptionPriceDeltaAmericanPutBinomial(spot, exercise, r, sigma, timeToMaturity, steps)
    print 'Call price partials'
    print 'delta, gamma, theta, vega, rho', OptionPricePartialsAmericanCallBinomial(spot, exercise, r, sigma, timeToMaturity, steps)
    print 'Put price partials'
    print 'delta, gamma, theta, vega, rho', OptionPricePartialsAmericanPutBinomial(spot, exercise, r, sigma, timeToMaturity, steps)









