#!/usr/bin/env python
##
# Copyright (c) 2000 by g.p.ciceri <gp.ciceri@acm.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# Alternatively, you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#                                   
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU (Library) General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# $Author: zanzi $
# $Revision: 1.1 $
# $Log: Bonds.py,v $
# Revision 1.1  2000/01/02 17:43:33  zanzi
# Initial revision
#
#
# module name: Bonds.py
#
# description: python functions to handle bond price algorithms
#
#

from types import ListType, TupleType
from math import exp

def BondsPriceSplit(couponTimes, couponAmounts, principalTimes, principalAmounts, rate):
    for theParameter in (couponTimes, couponAmounts, principalTimes, principalAmounts):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
        
    if not len(couponTimes) == len(couponAmounts):
        raise StandardError, "coupon times and amount must be of the same lenght"
    if not len(principalTimes) == len(principalAmounts):
        raise StandardError, "principal times and amount must be of the same lengh"
    
    p = 0.0
    for x in xrange(len(couponTimes)):
        p = p + exp(-rate * couponTimes[x]) * couponAmounts[x]
    for x in xrange(len(principalTimes)):
        p = p + exp(-rate * principalTimes[x]) * principalAmounts[x]

    return p

def BondsPrice(cflowTimes, cflowAmounts, rate):
    for theParameter in (cflowTimes, cflowAmounts):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
        
    if not len(cflowTimes) == len(cflowAmounts):
        raise StandardError, "cash flows times and amount must be of the same lenght"
    
    p = 0.0
    for x in xrange(len(cflowTimes)):
        p = p + exp(-rate * cflowTimes[x]) * cflowAmounts[x]
    return p


def BondsPriceDiscrete(cflowTimes, cflowAmounts, rate):
    for theParameter in (cflowTimes, cflowAmounts):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
        
    if not len(cflowTimes) == len(cflowAmounts):
        raise StandardError, "cash flows times and amount must be of the same lenght"
    
    p = 0.0
    for x in xrange(len(cflowTimes)):
        p = p + (cflowAmounts[x] / (pow((1+rate),cflowTimes[x])))
    return p

def BondsYieldToMaturity(cflowTimes, cflowAmounts, bondPrice):
    for theParameter in (cflowTimes, cflowAmounts):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
        
    if not len(cflowTimes) == len(cflowAmounts):
        raise StandardError, "cash flows times and amount must be of the same lenght"

    ACCURACY=1e-05
    MAX_ITERATIONS=200
    bot = 0.0
    top = 1.0
    while (BondsPrice(cflowTimes, cflowAmounts, top) > bondPrice):
        top = top * 2
    r = 0.5 * (top + bot)
    for x in xrange(MAX_ITERATIONS):
        diff = BondsPrice(cflowTimes, cflowAmounts, r) - bondPrice
        if (abs(diff) < ACCURACY): return r
        if (diff > 0.0): bot = r
        else: top = r
        r = 0.5 * (top + bot)
    return r

def BondsDuration(cflowTimes, cflowAmounts, rate):
    for theParameter in (cflowTimes, cflowAmounts):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
        
    if not len(cflowTimes) == len(cflowAmounts):
        raise StandardError, "cash flows times and amount must be of the same lenght"
    S = 0.0
    D1 = 0.0
    for x in xrange(len(cflowTimes)):
        S = S + (cflowAmounts[x] * exp ( -rate * cflowTimes[x]))
        D1 = D1 + (cflowTimes[x] * cflowAmounts[x] * exp ( -rate * cflowTimes[x]))
    return D1/S


def BondsDurationMacaulay(cflowTimes, cflowAmounts, bondPrice):
    for theParameter in (cflowTimes, cflowAmounts):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
        
    if not len(cflowTimes) == len(cflowAmounts):
        raise StandardError, "cash flows times and amount must be of the same lenght"
    y = BondsYieldToMaturity(cflowTimes, cflowAmounts, bondPrice)
    return BondsDuration(cflowTimes, cflowAmounts,y)
   
def BondsDurationModified(cflowTimes, cflowAmounts, bondPrice, rate):
    for theParameter in (cflowTimes, cflowAmounts):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
        
    if not len(cflowTimes) == len(cflowAmounts):
        raise StandardError, "cash flows times and amount must be of the same lenght"
    dur = BondsDuration(cflowTimes, cflowAmounts, rate)
    y = BondsYieldToMaturity(cflowTimes, cflowAmounts, bondPrice)
    return dur/y

def BondsConvexity(cflowTimes, cflowAmounts, yield):
    for theParameter in (cflowTimes, cflowAmounts):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
        
    if not len(cflowTimes) == len(cflowAmounts):
        raise StandardError, "cash flows times and amount must be of the same lenght"
    C = 0.0
    for x in xrange(len(cflowTimes)):
        t = cflowTimes[x]
        C = C + cflowAmounts[x] * t * t * exp (-y * t)
    return C



if __name__ == '__main__':
    coupontimes = [1, 2]
    couponamounts = [10, 10]
    principaltimes = [2]
    principalamounts = [100]
    rate = 0.1
    print 'Price:', BondsPriceSplit(coupontimes,couponamounts,principaltimes, principalamounts, rate)
    cflowtimes = [1, 2]
    cflowamounts = [10, 110]
    print 'Price, cashflow:', BondsPrice(cflowtimes,cflowamounts, rate)
    print 'Price, discrete:', BondsPriceDiscrete(cflowtimes,cflowamounts, rate)
    print 'Duration, simple:', BondsDuration(cflowtimes,cflowamounts, rate)
    price = BondsPrice(cflowtimes,cflowamounts, 0.11)
    print 'Duration, Macaulay:', BondsDurationMacaulay(cflowtimes,cflowamounts,price)
    print 'Duration, modified:', BondsDurationModified(cflowtimes,cflowamounts,price, rate)
    print 'Yield:', BondsYieldToMaturity(cflowtimes,cflowamounts, price)
    y = BondsYieldToMaturity(cflowtimes,cflowamounts, price)
    print 'Convexity:', BondsConvexity(cflowtimes,cflowamounts, y)











