#!/usr/bin/env python
##
# Copyright (c) 2000 by g.p.ciceri <gp.ciceri@acm.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# Alternatively, you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#                                   
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU (Library) General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.                         
#
#
# $Author: zanzi $
# $Revision: 1.2 $
# $Log: CashFlow.py,v $
# Revision 1.2  2000/01/07 18:36:17  zanzi
# *** empty log message ***
#
# Revision 1.1  2000/01/02 17:44:15  zanzi
# Initial revision
#
#
# module name: CashFlow.py
#
# description: python functions to handle cash flows and IRR
#
#

from types import ListType, TupleType
from math import exp

ERROR = -1.0e30

def CashFlowPV(cFlowTimes, cFlowAmounts, rate):
    for theParameter in (cFlowTimes, cFlowAmounts):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
    if not len(cFlowTimes) == len(cFlowAmounts):
        raise StandardError, "times and amount must be of the same length"
    
    PV = 0.0
    for x in xrange(len(cFlowTimes)):
        PV = PV + cFlowAmounts[x] * exp( -rate * cFlowTimes[x])
    return PV


def CashFlowPVDiscrete(cFlowTimes, cFlowAmounts, rate):
    for theParameter in (cFlowTimes, cFlowAmounts):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
    if not len(cFlowTimes) == len(cFlowAmounts):
        raise StandardError, "times and amount must be of the same length"
    
    PV = 0.0
    for x in xrange(len(cFlowTimes)):
        PV = PV + cFlowAmounts[x] / pow(1+rate, cFlowTimes[x])
    return PV

def CashFlowIRR(cFlowTimes, cFlowAmounts):
    for theParameter in (cFlowTimes, cFlowAmounts):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
    if not len(cFlowTimes) == len(cFlowAmounts):
        raise StandardError, "times and amount must be of the same length"

    ACCURACY = 1.0e-5
    MAX_ITERATIONS = 50
    x1 = 0.0
    x2 = 0.2

    f1 = CashFlowPV(cFlowTimes, cFlowAmounts, x1)
    f2 = CashFlowPV(cFlowTimes, cFlowAmounts, x2)

    for x in xrange(MAX_ITERATIONS):
        if ((f1 * f2) < 0.0): break
        if (abs(f1) < abs(f2)):
            x1 = x1 + (1.6 * (x1-x2))
            f1 = CashFlowPV(cFlowTimes, cFlowAmounts, x1)
        else:
            x2 = x2 + (1.6 * (x2-x1))
            f2 = CashFlowPV(cFlowTimes, cFlowAmounts, x2)
            
    if ((f2 * f1) > 0.0): return ERROR

    f = CashFlowPV(cFlowTimes, cFlowAmounts, x1)
    dx = 0
    if ( f < 0.0 ):
        rtb = x1
        dx = x2 - x1
    else:
        rtb = x2
        dx = x1 - x2
    for x in xrange(MAX_ITERATIONS):
        dx = dx * 0.5
        x_mid = rtb + dx
        f_mid = CashFlowPV(cFlowTimes, cFlowAmounts, x_mid)
        if ( f_mid <= 0.0 ): rtb = x_mid
        if ( (abs(f_mid) < ACCURACY) or (abs(dx) < ACCURACY) ): return x_mid
        
    return ERROR

def sgn(number):
    if (number >= 0): return 1
    else: return -1

def CashFlowUniqueIRR(cFlowTimes, cFlowAmounts):
    for theParameter in (cFlowTimes, cFlowAmounts):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
    if not len(cFlowTimes) == len(cFlowAmounts):
        raise StandardError, "times and amount must be of the same lenght"
    
    signChanges = 0
    for x in xrange(1,len(cFlowTimes)):
        if (sgn(cFlowAmounts[x-1])!= sgn(cFlowAmounts[x])): signChanges = signChanges + 1
    if signChanges == 0: return  0
    if signChanges == 1: return  1
    A = cFlowAmounts[0]
    B = 0
    signChanges = 0
    for x in xrange(1,len(cFlowTimes)):
        B = A + cFlowAmounts[x]
        if (sgn(A)!= sgn(B)): signChanges = signChanges + 1
        A = B
    if signChanges <= 1: return 1
    else: return 0




if __name__ == '__main__':
    times = [1, 2]
    amounts = [-100, 110.51709]
    print 'PV 0   (Continuous compounding):', CashFlowPV(times, amounts, 0.0)
    print 'PV 0.1 (Continuous compounding):', CashFlowPV(times, amounts, 0.1)
    print 'PV 0   (Discrete compounding)  :', CashFlowPVDiscrete(times, amounts, 0.0)
    print 'PV 0.1 (Discrete compounding)  :', CashFlowPVDiscrete(times, amounts, 0.1)
    amounts = [-100, 134.98588]
    print 'IRR (Continuous compounding):', CashFlowIRR(times, amounts)
    amounts = [100, -134.98588]
    print 'IRR (Continuous compounding):', CashFlowIRR(times, amounts)
    times.append(3)
    amounts.append(30)
    print 'IRR (Continuous compounding):', CashFlowIRR(times, amounts)
    amounts = [-100,110,10]
    print 'IRR unique:', CashFlowUniqueIRR(times, amounts)
    amounts = [-100,110,-10]
    print 'IRR unique:', CashFlowUniqueIRR(times, amounts)
    amounts = [-10,20.0001,-10]
    print 'IRR unique:', CashFlowUniqueIRR(times, amounts)
    print 'IRR (Continuous compounding):', CashFlowIRR(times, amounts)
    amounts = [-10,19.999,-10]
    print 'IRR unique:', CashFlowUniqueIRR(times, amounts)
    print 'IRR (Continuous compounding):', CashFlowIRR(times, amounts)
    amounts = [10,-21,10]
    print 'IRR unique:', CashFlowUniqueIRR(times, amounts)
    print 'IRR (Continuous compounding):', CashFlowIRR(times, amounts)
