#!/usr/bin/env python
##
# Copyright (c) 2000 by g.p.ciceri <gp.ciceri@acm.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# Alternatively, you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#                                   
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU (Library) General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.                    
#
#
# $Author: zanzi $
# $Revision: 1.1 $
# $Log: FiniteDiff.py,v $
# Revision 1.1  2000/01/07 18:35:38  zanzi
# Initial revision
#
#
#
# module name: FiniteDiff.py
#
# description: python functions to handle finite differences algorithms
#
#

from math import exp, sqrt, log, fmod
from NormDist import N
from Numeric import zeros, matrixmultiply, Float
from LinearAlgebra import inverse
from BasicOptions import OptionPricePutBlackScholes, OptionPriceCallBlackScholes

def OptionPricePutEuropeanFiniteDiffImplicit(spotPrice, strikePrice, interestRate, sigma, time, stepSNum, stepTNum):
    sigmaSqr = sigma * sigma
    if (fmod(stepSNum,2) == 1): M = stepSNum + 1
    else: M = stepSNum
    deltaS = 2.0 * spotPrice / M
    spotValues = []
    for m in xrange(M):
        spotValues.append(m * deltaS)
    deltaT = time/stepTNum
    A = zeros((M+1,M+1), Float)
    A[0,0] = 1.0
    for j in xrange(1,M):
        A[j][j-1]   = 0.5 * j * deltaT * (r - sigmaSqr * j)
        A[j][j]   = 1.0 + deltaT * (r + sigmaSqr * j * j)
        A[j][j+1]     = 0.5 * j * deltaT * (-r - sigmaSqr * j)
    A[M][M] = 1.0
    B = zeros(M+1, Float)
    for j in xrange(M):
        B[j] = max(0.0, strikePrice - spotValues[j])
    F = matrixmultiply(inverse(A),B)
    for t in xrange(stepTNum - 1, 0, -1):
        B = F
        F = matrixmultiply(inverse(A),B)
    return F[M/2]

def OptionPriceCallEuropeanFiniteDiffImplicit(spotPrice, strikePrice, interestRate, sigma, time, stepSNum, stepTNum):
    sigmaSqr = sigma * sigma
    if (fmod(stepSNum,2) == 1): M = stepSNum + 1
    else: M = stepSNum
    deltaS = 2.0 * spotPrice / M
    spotValues = []
    for m in xrange(M):
        spotValues.append(m * deltaS)
    deltaT = time/stepTNum
    A = zeros((M+1,M+1), Float)
    A[0,0] = 1.0
    for j in xrange(1,M):
        A[j][j-1]   = 0.5 * j * deltaT * (r - sigmaSqr * j)
        A[j][j]   = 1.0 + deltaT * (r + sigmaSqr * j * j)
        A[j][j+1]     = 0.5 * j * deltaT * (-r - sigmaSqr * j)
    A[M][M] = 1.0
    B = zeros(M+1, Float)
    for j in xrange(M):
        B[j] = max(0.0, - strikePrice + spotValues[j])
    F = matrixmultiply(inverse(A),B)
    for t in xrange(stepTNum - 1, 0, -1):
        B = F
        F = matrixmultiply(inverse(A),B)
    return F[M/2]

def OptionPricePutEuropeanFiniteDiffExplicit(spotPrice, strikePrice, interestRate, sigma, time, stepSNum, stepTNum):
    sigmaSqr = sigma * sigma
    if (fmod(stepSNum,2) == 1): M = stepSNum + 1
    else: M = stepSNum
    deltaS = 2.0 * spotPrice / M
    spotValues = []
    for m in xrange(M):
        spotValues.append(m * deltaS)
    deltaT = time/stepTNum

    a = zeros(M, Float)
    b = zeros(M, Float)
    c = zeros(M, Float)
    r1 = 1.0/(1.0 + r * deltaT)
    r2 = deltaT / (1.0 + r * deltaT)
    
    for j in xrange(1,M):
        a[j]   = r2 * 0.5 * j * (- r + sigmaSqr * j)
        b[j]   = r1 * (1.0 - sigmaSqr * j * j * deltaT)
        c[j]   = r2 * 0.5 * j * (r + sigmaSqr * j)
 
    fNext = zeros(M+1, Float)
    f = zeros(M+1, Float)
    
    for j in xrange(M):
        fNext[j] = max(0.0, strikePrice - spotValues[j])
        
    for t in xrange(stepTNum - 1, -1, -1):
        f[0] = X
        for m in xrange(1,M):
            f[m] = a[m] * fNext[m-1] + b[m] * fNext[m] + c[m] * fNext[m+1]
        f[M] = 0
        for m in xrange(M-1):
            fNext[m] = f[m]
    return f[M/2]

def OptionPriceCallEuropeanFiniteDiffExplicit(spotPrice, strikePrice, interestRate, sigma, time, stepSNum, stepTNum):
    sigmaSqr = sigma * sigma
    if (fmod(stepSNum,2) == 1): M = stepSNum + 1
    else: M = stepSNum
    deltaS = 2.0 * spotPrice / M
    spotValues = []
    for m in xrange(M):
        spotValues.append(m * deltaS)
    deltaT = time/stepTNum

    a = zeros(M, Float)
    b = zeros(M, Float)
    c = zeros(M, Float)
    r1 = 1.0/(1.0 + r * deltaT)
    r2 = deltaT / (1.0 + r * deltaT)
    
    for j in xrange(1,M):
        a[j]   = r2 * 0.5 * j * (- r + sigmaSqr * j)
        b[j]   = r1 * (1.0 - sigmaSqr * j * j * deltaT)
        c[j]   = r2 * 0.5 * j * (r + sigmaSqr * j)
 
    fNext = zeros(M+1, Float)
    f = zeros(M+1, Float)
    
    for j in xrange(M):
        fNext[j] = max(0.0, - strikePrice + spotValues[j])
        
    for t in xrange(stepTNum - 1, -1, -1):
        f[0] = X
        for m in xrange(1,M):
            f[m] = a[m] * fNext[m-1] + b[m] * fNext[m] + c[m] * fNext[m+1]
        f[M] = 0
        for m in xrange(M-1):
            fNext[m] = f[m]
    return f[M/2]

def OptionPricePutAmericanFiniteDiffImplicit(spotPrice, strikePrice, interestRate, sigma, time, stepSNum, stepTNum):
    sigmaSqr = sigma * sigma
    if (fmod(stepSNum,2) == 1): M = stepSNum + 1
    else: M = stepSNum
    deltaS = 2.0 * spotPrice / M
    spotValues = []
    for m in xrange(M+1):
        spotValues.append(m * deltaS)
    deltaT = time/stepTNum
    A = zeros((M+1,M+1), Float)
    A[0,0] = 1.0
    for j in xrange(1,M):
        A[j][j-1]   = 0.5 * j * deltaT * (r - sigmaSqr * j)
        A[j][j]   = 1.0 + deltaT * (r + sigmaSqr * j * j)
        A[j][j+1]     = 0.5 * j * deltaT * (-r - sigmaSqr * j)
    A[M][M] = 1.0
    B = zeros(M+1, Float)
    for j in xrange(M+1):
        B[j] = max(0.0, strikePrice - spotValues[j])
    F = matrixmultiply(inverse(A),B)
    for t in xrange(stepTNum - 1, 0, -1):
        B = F
        F = matrixmultiply(inverse(A),B)
        for m in xrange(1,M-1):
            F[m] = max(F[m], strikePrice - spotValues[m])
    return F[M/2]

def OptionPriceCallAmericanFiniteDiffImplicit(spotPrice, strikePrice, interestRate, sigma, time, stepSNum, stepTNum):
    sigmaSqr = sigma * sigma
    if (fmod(stepSNum,2) == 1): M = stepSNum + 1
    else: M = stepSNum
    deltaS = 2.0 * spotPrice / M
    spotValues = []
    for m in xrange(M+1):
        spotValues.append(m * deltaS)
    deltaT = time/stepTNum
    A = zeros((M+1,M+1), Float)
    A[0,0] = 1.0
    for j in xrange(1,M):
        A[j][j-1]   = 0.5 * j * deltaT * (r - sigmaSqr * j)
        A[j][j]   = 1.0 + deltaT * (r + sigmaSqr * j * j)
        A[j][j+1]     = 0.5 * j * deltaT * (-r - sigmaSqr * j)
    A[M][M] = 1.0
    B = zeros(M+1, Float)
    for j in xrange(M+1):
        B[j] = max(0.0, - strikePrice + spotValues[j])
    F = matrixmultiply(inverse(A),B)
    for t in xrange(stepTNum - 1, 0, -1):
        B = F
        F = matrixmultiply(inverse(A),B)
        for m in xrange(1,M-1):
            F[m] = max(F[m], - strikePrice + spotValues[m])
    return F[M/2]

def OptionPricePutAmericanFiniteDiffExplicit(spotPrice, strikePrice, interestRate, sigma, time, stepSNum, stepTNum):
    sigmaSqr = sigma * sigma
    if (fmod(stepSNum,2) == 1): M = stepSNum + 1
    else: M = stepSNum
    deltaS = 2.0 * spotPrice / M
    spotValues = []
    for m in xrange(M+1):
        spotValues.append(m * deltaS)
    deltaT = time/stepTNum

    a = zeros(M, Float)
    b = zeros(M, Float)
    c = zeros(M, Float)
    r1 = 1.0/(1.0 + r * deltaT)
    r2 = deltaT / (1.0 + r * deltaT)
    
    for j in xrange(1,M):
        a[j]   = r2 * 0.5 * j * (- r + sigmaSqr * j)
        b[j]   = r1 * (1.0 - sigmaSqr * j * j * deltaT)
        c[j]   = r2 * 0.5 * j * (r + sigmaSqr * j)
 
    fNext = zeros(M+1, Float)
    f = zeros(M+1, Float)
    
    for j in xrange(M+1):
        fNext[j] = max(0.0, strikePrice - spotValues[j])
        
    for t in xrange(stepTNum - 1, -1, -1):
        f[0] = 0
        for m in xrange(1,M):
            f[m] = a[m] * fNext[m-1] + b[m] * fNext[m] + c[m] * fNext[m+1]
            f[m] = max(f[m], strikePrice - spotValues[m])
        f[M] = strikePrice - spotValues[M]
        for m in xrange(M):
            fNext[m] = f[m]
    return f[M/2]

def OptionPriceCallAmericanFiniteDiffExplicit(spotPrice, strikePrice, interestRate, sigma, time, stepSNum, stepTNum):
    sigmaSqr = sigma * sigma
    if (fmod(stepSNum,2) == 1): M = stepSNum + 1
    else: M = stepSNum
    deltaS = 2.0 * spotPrice / M
    spotValues = []
    for m in xrange(M+1):
        spotValues.append(m * deltaS)
    deltaT = time/stepTNum

    a = zeros(M, Float)
    b = zeros(M, Float)
    c = zeros(M, Float)
    r1 = 1.0/(1.0 + r * deltaT)
    r2 = deltaT / (1.0 + r * deltaT)
    
    for j in xrange(1,M):
        a[j]   = r2 * 0.5 * j * (- r + sigmaSqr * j)
        b[j]   = r1 * (1.0 - sigmaSqr * j * j * deltaT)
        c[j]   = r2 * 0.5 * j * (r + sigmaSqr * j)
 
    fNext = zeros(M+1, Float)
    f = zeros(M+1, Float)
    
    for j in xrange(M+1):
        fNext[j] = max(0.0, - strikePrice + spotValues[j])
        
    for t in xrange(stepTNum - 1, -1, -1):
        f[0] = 0
        for m in xrange(1,M):
            f[m] = a[m] * fNext[m-1] + b[m] * fNext[m] + c[m] * fNext[m+1]
            f[m] = max(f[m], - strikePrice + spotValues[m])
        f[M] = - strikePrice + spotValues[M]
        for m in xrange(M):
            fNext[m] = f[m]
    return f[M/2]

if __name__ == '__main__':
    S = 50.0
    X = 50.0
    r = 0.1
    sigma = 0.4
    timeToMaturity = 0.4167
    noSsteps = 20
    noTsteps = 11
    print 'black scholes call price ', OptionPriceCallBlackScholes(S, X, r, sigma, timeToMaturity)
    print 'explicit FD American call price', OptionPriceCallAmericanFiniteDiffExplicit(S, X, r, sigma, timeToMaturity, noSsteps, noTsteps)
    print 'implicit FD American call price', OptionPriceCallAmericanFiniteDiffImplicit(S, X, r, sigma, timeToMaturity, noSsteps, noTsteps)
    print 'black scholes put price ', OptionPricePutBlackScholes(S, X, r, sigma, timeToMaturity)   
    print 'explicit FD Euro put price', OptionPricePutEuropeanFiniteDiffExplicit(S, X, r, sigma, timeToMaturity, noSsteps, noTsteps)
    print 'implicit FD Euro put price', OptionPricePutEuropeanFiniteDiffImplicit(S, X, r, sigma, timeToMaturity, noSsteps, noTsteps)
    print 'explicit FD American put price', OptionPricePutAmericanFiniteDiffExplicit(S, X, r, sigma, timeToMaturity, noSsteps, noTsteps)
    print 'implicit FD American put price', OptionPricePutAmericanFiniteDiffImplicit(S, X, r, sigma, timeToMaturity, noSsteps, noTsteps)





