#!/usr/bin/env python
##
# Copyright (c) 2000 by g.p.ciceri <gp.ciceri@acm.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# Alternatively, you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#                                   
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU (Library) General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.                    
#
#
# $Author: zanzi $
# $Revision: 1.1 $
# $Log: FixedIncome.py,v $
# Revision 1.1  2000/01/07 18:36:01  zanzi
# Initial revision
#
#
#
# module name: FixedIncome.py
#
# description: python functions to handle fixed income modelling algorithms
#
#

from math import exp, sqrt, log
from NormDist import N
from types import ListType, TupleType
import TermStructure

def BondOptionPriceCallZeroBlackScholes(B, X, r, sigma, time):
    timeSqrt = sqrt(time)
    d1 = (log(B/X)+r*time)/(sigma*timeSqrt) + 0.5*sigma*timeSqrt
    d2 = d1-(sigma*timeSqrt)
    c = B * N(d1) - X * exp(-r*time) * N(d2)
    return c

def BondOptionPricePutZeroBlackScholes(B, X, r, sigma, time):
    timeSqrt = sqrt(time)
    d1 = (log(B/X)+r*time)/(sigma*timeSqrt) + 0.5*sigma*timeSqrt
    d2 = d1-(sigma*timeSqrt)
    c = - B * N(-d1) + X * exp(-r*time) * N(-d2)
    return c

def BondOptionPriceCallCouponBondBlackScholes(B, X, r, sigma, time, couponTimes, couponAmounts):
    for theParameter in (couponTimes, couponAmounts):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
    if not len(couponTimes) == len(couponAmounts):
        raise StandardError, "times and amounts must be of the same length"

    for i in xrange(len(couponTimes)):
        if (couponTimes[i]<=time):
            B = B - couponAmounts[i] * exp(-r*couponTimes[i])
    return BondOptionPriceCallZeroBlackScholes(B,X,r,sigma,time)

def BondOptionPricePutCouponBondBlackScholes(B, X, r, sigma, time, couponTimes, couponAmounts):
    for theParameter in (couponTimes, couponAmounts):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
    if not len(couponTimes) == len(couponAmounts):
        raise StandardError, "times and amounts must be of the same length"

    for i in xrange(len(couponTimes)):
        if (couponTimes[i]<=time):
            B = B - couponAmounts[i] * exp(-r*couponTimes[i])
    return BondOptionPricePutZeroBlackScholes(B,X,r,sigma,time)

def BondOptionPriceCallZeroAmericanRendlemanBartter(X, optionMaturity, S, M, interest, bondMaturity, maturityPayment, stepsNum):

    deltaT = bondMaturity/stepsNum
    u=exp(S*sqrt(deltaT))
    d=1.0/u
    pUp = (exp(M*deltaT)-d)/(u-d)
    pDown = 1.0-pUp
    r = []
    r.append(interest*pow(d,stepsNum))
    uu=u*u
    for i in xrange(1,stepsNum+1):
        r.append(r[i-1]*uu)
    P = []
    for i in xrange(stepsNum+1):
        P.append(maturityPayment)
    callStepsNum=int(stepsNum*optionMaturity/bondMaturity)
    for currStep in xrange(stepsNum, callStepsNum, - 1):
        for i in xrange(currStep):
            r[i] = r[i]*u;
            P[i] = exp(-r[i]*deltaT)*(pDown*P[i]+pUp*P[i+1])
    C = []
    for i in xrange(0,callStepsNum+1):
        C.append(max(0.0,P[i]-X))
    for currStep in xrange(callStepsNum, -1, - 1):
        for i in xrange(currStep):
            r[i] = r[i]*u;
            P[i] = exp(-r[i]*deltaT)*(pDown*P[i]+pUp*P[i+1])
            C[i]= max(P[i]-X, exp(-r[i]*deltaT)*(pUp*C[i+1]+pDown*C[i]))
    return C[0]   

def BondOptionPriceCallZeroVasicek(X, r, optionTimeToMaturity, bondTimeToMaturity, a, b, sigma):
    Tt = optionTimeToMaturity
    st = bondTimeToMaturity
    Ts = st-Tt
    if (a==0.0):
        vtT = sigma * sqrt(Tt)
        sigmaP = sigma*Ts*sqrt(Tt)
    else:
        vtT = sqrt (sigma*sigma*(1-exp(-2*a*Tt))/(2*a))
        BTs = (1-exp(-a*Ts))/a
        sigmaP = vtT*BTs
    h = (1.0/sigmaP) * log (TermStructure.TermStructureDiscountFactorVasicek(st,r,a,b,sigma)/(TermStructure.TermStructureDiscountFactorVasicek(Tt,r,a,b,sigma)*X) ) + sigmaP/2.0
    c = TermStructure.TermStructureDiscountFactorVasicek(st,r,a,b,sigma)*N(h) -X*TermStructure.TermStructureDiscountFactorVasicek(Tt,r,a,b,sigma)*N(h-sigmaP)
    return c

def BondOptionPricePutZeroVasicek(X, r, optionTimeToMaturity, bondTimeToMaturity, a, b, sigma):
    Tt = optionTimeToMaturity
    st = bondTimeToMaturity
    Ts = st-Tt
    if (a==0.0):
        vtT = sigma * sqrt(Tt)
        sigmaP = sigma*Ts*sqrt(Tt)
    else:
        vtT = sqrt (sigma*sigma*(1-exp(-2*a*Tt))/(2*a))
        BTs = (1-exp(-a*Ts))/a
        sigmaP = vtT*BTs
    h = (1.0/sigmaP) * log (TermStructure.TermStructureDiscountFactorVasicek(st,r,a,b,sigma)/(TermStructure.TermStructureDiscountFactorVasicek(Tt,r,a,b,sigma)*X) ) + sigmaP/2.0
    c = - TermStructure.TermStructureDiscountFactorVasicek(st,r,a,b,sigma)*N(-h) + TermStructure.TermStructureDiscountFactorVasicek(Tt,r,a,b,sigma)*N(-h+sigmaP)
    return c  


if __name__ == '__main__':
    B=100
    X=100
    r=0.05
    sigma=0.1
    time=1
    print 'Black Scholes Bond options'
    print 'call ', BondOptionPriceCallZeroBlackScholes(B,X,r,sigma,time)
    print 'put ', BondOptionPricePutZeroBlackScholes(B,X,r,sigma,time)
    coupon = (1,)
    couponTimes = (0.5,)
    print 'price adjusting for coupon'
    print 'call ', BondOptionPriceCallCouponBondBlackScholes(B,X,r,sigma,time,couponTimes,coupon)
    print 'put ', BondOptionPricePutCouponBondBlackScholes(B,X,r,sigma,time,couponTimes,coupon)

    print 'Rendleman and Bartter: Zero Coupon Call '
    X=1000
    S=0.15
    M=0.05
    interest=0.10
    no_steps=100
    option_maturity=4.0
    bond_maturity=5.0
    maturity_payment=1000
    X=900
    print 'option on zero coupon bond'
    print BondOptionPriceCallZeroAmericanRendlemanBartter( X, option_maturity, S, M, interest, bond_maturity, maturity_payment, no_steps)
    print 'testing Bond option, Vasicek'
    a = 0.1
    b = 0.1
    sigma = 0.02
    r = 0.05
    print ' discount factor, t=1: ', TermStructure.TermStructureDiscountFactorVasicek(1.0,r,a,b,sigma)
    X=0.9
    print 'call option price ', BondOptionPriceCallZeroVasicek(X,r,1,5,a,b,sigma)
    print 'put option price ', BondOptionPricePutZeroVasicek(X,r,1,5,a,b,sigma)   
    


