#!/usr/bin/env python
##
# Copyright (c) 2000 by g.p.ciceri <gp.ciceri@acm.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# Alternatively, you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#                                   
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU (Library) General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.                    
#
#
# $Author: zanzi $
# $Revision: 1.1 $
# $Log: GeneralSimulation.py,v $
# Revision 1.1  2000/01/07 18:36:49  zanzi
# Initial revision
#
# Revision 1.1  2000/01/02 17:44:08  zanzi
# Initial revision
#
#
# module name: GeneralSimulation.py
#
# description: python functions to handle general simulation prices algorithms
#
#

from math import exp, sqrt, log
from Random import RandomNormal
import BasicOptions

def SimulateTerminalPrice(spotPrice, interestRate, sigma, time, seed):
    sigmaSqr = sigma * sigma
    R = (interestRate - 0.5 * sigmaSqr) * time
    SD = sigma * sqrt(time)
    sumPayoffs = 0.0
    RV = [0, seed] # RV[1] is the seed
    RV = RandomNormal(RV[1])
    return spotPrice * exp(R + SD * RV[0]), RV[1]                

def SimulatePriceSequence(spotPrice, interestRate, sigma, time, stepsNum, seed):
    prices = []
    deltaT = time/stepsNum
    sigmaSqr = sigma * sigma
    R = (interestRate - 0.5 * sigmaSqr) * deltaT
    SD = sigma * sqrt(deltaT)
    St = spotPrice
    RV = [0, seed]
    for m in xrange(stepsNum):
        RV = RandomNormal(RV[1])
        St = St * exp(R + SD * RV[0]) 
        prices.append(St)
    return prices, RV[1]

def PayoffEuropeanCall(price, strikePrice):
    return max(0.0, price - strikePrice)

def PayoffEuropeanPut(price, strikePrice):
    return max(0.0, - price + strikePrice)

def PayoffMax(prices):
    m = max(prices)
    return m - prices[-1]

def PayoffMin(prices):
    m = min(prices)
    return prices[-1] - m

def PayoffArithmeticAverage(prices):
    sum = 0.0
    for i in xrange(len(prices)):
        sum = sum + prices[i]
    avg = sum / len(prices)
    return max(0.0, avg - prices[-1])

def PayoffGeometricAverage(prices):
    logsum = 0.0
    for i in xrange(len(prices)):
        logsum = logsum + log(prices[i])
    avg = exp(logsum / len(prices))
    return max(0.0, avg - prices[-1])

def DerivativePriceEuropeanSimulated(spotPrice, strikePrice, interestRate, sigma, time, payoff, simNum, seed):
    sumPayoffs = 0;
    for i in xrange(simNum):
        (ST, seed) = SimulateTerminalPrice(spotPrice, interestRate, sigma, time, seed)
        sumPayoffs = sumPayoffs + payoff(ST, strikePrice)

    return exp(- interestRate * time) * (sumPayoffs / simNum)

# with Vector of Prices, without strikePrice
def DerivativePriceEuropeanSimulatedVP(spotPrice, interestRate, sigma, time, payoff, stepNum, simNum, seed):
    sumPayoffs = 0.0;
    for i in xrange(simNum):
        (prices, seed) = SimulatePriceSequence(spotPrice, interestRate, sigma, time, stepNum, seed)
        sumPayoffs = sumPayoffs + payoff(prices)

    return exp(- interestRate * time) * (sumPayoffs / simNum)

# with Vector of Prices and strikePrice
def DerivativePriceEuropeanSimulatedVPX(spotPrice, interestRate, sigma, time, payoff, strikePrice, stepNum, simNum, seed):
    sumPayoffs = 0.0;
    for i in xrange(simNum):
        (prices, seed) = SimulatePriceSequence(spotPrice, interestRate, sigma, time, stepNum, seed)
        sumPayoffs = sumPayoffs + payoff(prices, strikePrice)

    return exp(- interestRate * time) * (sumPayoffs / simNum)


# with Vector of Prices, control variate
def DerivativePriceEuropeanSimulatedVPControlVariate(spotPrice, interestRate, sigma, time, payoff, stepNum, simNum, seed):
    strikePrice = spotPrice
    cBS = BasicOptions.OptionPriceCallBlackScholes(spotPrice, strikePrice, interestRate, sigma, time)
    sumPayoffs = 0.0;
    sumPayoffsBS = 0.0;
    for i in xrange(simNum):
        (prices, seed) = SimulatePriceSequence(spotPrice, interestRate, sigma, time, stepNum, seed)
        sumPayoffs = sumPayoffs + payoff(prices)
        sumPayoffsBS = sumPayoffsBS + PayoffEuropeanCall(prices[-1], strikePrice)
    cSim = exp(- interestRate * time) * (sumPayoffs / simNum)
    cBSSim = exp(- interestRate * time) * (sumPayoffsBS / simNum)
    cSim = cSim + (cBS - cBSSim)
    return cSim

if __name__ == '__main__':
    spot = 100.0
    exercise = 100.0
    r = 0.05
    sigma = 0.2
    timeToMaturity = 1.0
    simsNum = 10000
    seed = 1L
    print 'black scholes call price =', BasicOptions.OptionPriceCallBlackScholes(spot, exercise, r, sigma, timeToMaturity)
    print 'black scholes price call (simulated)=', DerivativePriceEuropeanSimulated(spot, exercise, r, sigma, timeToMaturity, PayoffEuropeanCall, simsNum, seed)
    print 'black scholes put price =', BasicOptions.OptionPricePutBlackScholes(spot, exercise, r, sigma, timeToMaturity)
    print 'black scholes price put (simulated)=', DerivativePriceEuropeanSimulated(spot, exercise, r, sigma, timeToMaturity, PayoffEuropeanPut, simsNum, seed)
    simsNum = 500
    stepNum = 300
    print 'simulated arithmetic average'
    print 'S = ' , spot, ' r= ', r, ' price =',
    print DerivativePriceEuropeanSimulatedVP(spot, r, sigma, timeToMaturity, PayoffArithmeticAverage, stepNum, simsNum, seed)
    print 'simulated geometric average'
    print 'price =', DerivativePriceEuropeanSimulatedVP(spot, r, sigma, timeToMaturity, PayoffGeometricAverage, stepNum, simsNum, seed)
    print 'simulated geometric average, control variate'
    print 'price =', DerivativePriceEuropeanSimulatedVPControlVariate(spot, r, sigma, timeToMaturity, PayoffGeometricAverage, stepNum, simsNum, seed)
    print 'simulated max'
    print 'price =', DerivativePriceEuropeanSimulatedVP(spot, r, sigma, timeToMaturity, PayoffMax, stepNum, simsNum, seed)
    print 'simulated min'
    print 'price =', DerivativePriceEuropeanSimulatedVP(spot, r, sigma, timeToMaturity, PayoffMin, stepNum, simsNum, seed)


  





