#!/usr/bin/env python
##
# Copyright (c) 2000 by g.p.ciceri <gp.ciceri@acm.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# Alternatively, you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#                                   
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU (Library) General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.                    
#
#
# $Author: zanzi $
# $Revision: 1.1 $
# $Log: Simulation.py,v $
# Revision 1.1  2000/01/02 17:44:08  zanzi
# Initial revision
#
#
# module name: Simulation.py
#
# description: python functions to handle simulation prices algorithms
#
#

from math import exp, sqrt, log
from Random import RandomNormal
import BasicOptions

def OptionPriceCallEuropeanSimulated(spotPrice, strikePrice, interestRate, sigma, time, simsNum):

    sigmaSqr = sigma * sigma
    R = (interestRate - 0.5 * sigmaSqr) * time
    SD = sigma * sqrt(time)
    sumPayoffs = 0.0
    RV = [0, 1L] # RV[1] is the seed
    for i in xrange(1, simsNum+1):
        RV = RandomNormal(RV[1])
        S_T = spotPrice * exp(R + SD * RV[0])
        sumPayoffs = sumPayoffs + max(0.0, S_T - strikePrice)
    c = exp(- interestRate * time) * (sumPayoffs/simsNum)
    return c

def OptionPricePutEuropeanSimulated(spotPrice, strikePrice, interestRate, sigma, time, simsNum):

    sigmaSqr = sigma * sigma
    R = (interestRate - 0.5 * sigmaSqr) * time
    SD = sigma * sqrt(time)
    sumPayoffs = 0.0
    RV = [0, 1L] # RV[1] is the seed
    for i in xrange(1, simsNum+1):
        RV = RandomNormal(RV[1])
        S_T = spotPrice * exp(R + SD * RV[0])
        sumPayoffs = sumPayoffs + max(0.0, - S_T + strikePrice)
    c = exp(- interestRate * time) * (sumPayoffs/simsNum)
    return c

def OptionPriceDeltaCallEuropeanSimulated(spotPrice, strikePrice, interestRate, sigma, time, simsNum):

    sigmaSqr = sigma * sigma
    R = (interestRate - 0.5 * sigmaSqr) * time
    SD = sigma * sqrt(time)
    sumPayoffs = 0.0
    sumPayoffs2 = 0.0
    q = spotPrice * 0.01
    RV = [0, 1L] # RV[1] is the seed
    for i in xrange(1, simsNum+1):
        RV = RandomNormal(RV[1])
        S_T = spotPrice * exp(R + SD * RV[0])
        sumPayoffs = sumPayoffs + max(0.0, S_T - strikePrice)
        S_T2 = (spotPrice + q) * exp(R + SD * RV[0])
        sumPayoffs2 = sumPayoffs2 + max(0.0, S_T2 - strikePrice)
         
    c = exp(- interestRate * time) * (sumPayoffs/simsNum)
    c2 = exp(- interestRate * time) * (sumPayoffs2/simsNum)
    
    return (c2 - c)/q

def OptionPriceDeltaPutEuropeanSimulated(spotPrice, strikePrice, interestRate, sigma, time, simsNum):

    sigmaSqr = sigma * sigma
    R = (interestRate - 0.5 * sigmaSqr) * time
    SD = sigma * sqrt(time)
    sumPayoffs = 0.0
    sumPayoffs2= 0.0
    q = spotPrice * 0.01
    RV = [0, 1L] # RV[1] is the seed
    for i in xrange(1, simsNum+1):
        RV = RandomNormal(RV[1])
        S_T = spotPrice * exp(R + SD * RV[0])
        sumPayoffs = sumPayoffs + max(0.0, - S_T + strikePrice)
        S_T2 = (spotPrice + q )* exp(R + SD * RV[0])
        sumPayoffs2 = sumPayoffs2 + max(0.0, - S_T2 + strikePrice)
         
    c = exp(- interestRate * time) * (sumPayoffs/simsNum)
    c2 = exp(- interestRate * time) * (sumPayoffs2/simsNum)
    
    return (c2 - c)/q


if __name__ == '__main__':
    spot = 100.0
    exercise = 100.0
    r = 0.1
    sigma = 0.25
    timeToMaturity = 1.0
    simsNum = 50000
    print 'call'
    print 'black scholes price =', BasicOptions.OptionPriceCallBlackScholes(spot, exercise, r, sigma, timeToMaturity)
    print 'black scholes price (simulated)=', OptionPriceCallEuropeanSimulated(spot, exercise, r, sigma, timeToMaturity, simsNum)

    print 'put' 
    print 'black scholes price =', BasicOptions.OptionPricePutBlackScholes(spot, exercise, r, sigma, timeToMaturity)
    print 'black scholes price (simulated)=', OptionPricePutEuropeanSimulated(spot, exercise, r, sigma, timeToMaturity, simsNum)
   
    print 'call'
    print 'black scholes delta =', BasicOptions.OptionPriceDeltaCallBlackScholes(spot, exercise, r, sigma, timeToMaturity)
    print 'black scholes delta (simulated)=', OptionPriceDeltaCallEuropeanSimulated(spot, exercise, r, sigma, timeToMaturity, simsNum)

    print 'put'
    print 'black scholes delta =', BasicOptions.OptionPriceDeltaPutBlackScholes(spot, exercise, r, sigma, timeToMaturity)
    print 'black scholes delta (simulated)=', OptionPriceDeltaPutEuropeanSimulated(spot, exercise, r, sigma, timeToMaturity, simsNum)







