#!/usr/bin/env python
##
# Copyright (c) 2000 by g.p.ciceri <gp.ciceri@acm.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# Alternatively, you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#                                   
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU (Library) General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.                    
#
#
# $Author: zanzi $
# $Revision: 1.1 $
# $Log: TermStructure.py,v $
# Revision 1.1  2000/01/07 18:37:12  zanzi
# Initial revision
#
#
#
# module name: TermStructure.py
#
# description: python functions to handle term structure algorithms
#
#

from math import exp, sqrt, log
from Random import RandomNormal
from types import ListType, TupleType

def TermStructureYieldFromDiscountFactor(dfact, t):
    return (-log(dfact)/t)

def TermStructureDiscountFactorFromYield(r, t):
    return (exp(-r)/t)

def TermStructureForwardRateFromDiscountFactor(dt, dT, t):
    return (log(dt/dT)/t)

def TermStructureForwardRateFromYield(rt1, rT, t1, T):
    return (rT * (T/(T-t1)) - rt1 * (t1/T))


def TermStructureYieldLinearlyInterpolated(time, obsTimes, obsYields):
    for theParameter in (obsTimes, obsYields):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
    if not len(obsTimes) == len(obsYields):
        raise StandardError, "times and yields must be of the same length"
    noObs = len(obsTimes)
    if noObs < 1: return 0
    tMin = obsTimes[0]
    if time <= tMin: return obsYields[0]
    tMax = obsTimes[-1]
    if time >= tMax: return obsYields[-1]
    t = 1
    while( (t < noObs) and (time > obsTimes[t])): t = t + 1
    Lambda = (obsTimes[t] - time) / (obsTimes[t] - obsTimes[t-1])
    r = obsYields[t-1] * Lambda + obsYields[t] * (1 - Lambda)
    return r

def TermStructureForwardLinearlyInterpolated(time, obsTimes, obsForwards):
    for theParameter in (obsTimes, obsForwards):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
    if not len(obsTimes) == len(obsForwards):
        raise StandardError, "times and forwards must be of the same length"
    noObs = len(obsTimes)
    if noObs < 1: return 0
    tMin = obsTimes[0]
    if time <= tMin: return obsForwards[0]
    tMax = obsTimes[-1]
    if time >= tMax: return obsForwards[-1]
    t = 1
    while( (t < noObs) and (time > obsTimes[t])): t = t + 1
    Lambda = (obsTimes[t] - time) / (obsTimes[t] - obsTimes[t-1])
    r = obsForwards[t-1] * Lambda + obsForwards[t] * (1 - Lambda)
    return r

def TermStructureYieldBliss(t, gamma0, gamma1, gamma2, lambda1, lambda2):
    t1 = t/lambda1
    t2 = t/lambda2
    r = gamma0 + gamma1 * ( (1-exp(-t1)) / t1 ) + gamma2 * ( ( 1-exp(-t2)) / t2 ) + gamma2 * ( -exp(-t2))
    return r

def TermStructureYieldNelsonSiegel(t, beta0, beta1, beta2, Lambda):
    if (t==0.0): return beta0
    tl = t/Lambda;
    r = beta0 + (beta1+beta2) * ((1-exp(-tl))/tl) + beta2 * exp(-tl);
    return r;     

def TermStructureDiscountFactorCubicSpline(t, b1, c1, d1, f, knots):
    for theParameter in (f, knots):
        if not (isinstance(theParameter, ListType) or (isinstance(theParameter, TupleType))):
            raise TypeError, "this parameter must be a tuple or a list"
    if not len(f) == len(knots):
        raise StandardError, "f and knots must be of the same length"

    d = 1.0 + b1 * t + c1 * (pow(t,2)) + d1 * (pow(t,3))
    for i in xrange(len(knots)):
        if (t >= knots[i]): d = d + f[i] * (pow((t-knots[i]),3))
        else: break
    return d

def TermStructureDiscountFactorVasicek(time, r, a, b, sigma):
    sigmaSqr = sigma * sigma
    aa = a * a
    if (aa == 0):
        B = time
        A = exp(sigmaSqr * pow(time,3))/6.0
    else:
        B = (1.0 - exp(-a * time))/a
        A = exp(((B-time) * (aa*b - 0.5 * sigmaSqr))/aa - ((sigmaSqr*B*B)/(4*a)))
    d = A * exp(-B*r)
    return d

def TermStructureDiscountFactorCir(t, r, kappa, Lambda, theta, sigma):
    sigmaSqr = sigma * sigma
    gamma = sqrt(pow((kappa + Lambda),2) + 2.0 * sigmaSqr)
    denum = (gamma + kappa + Lambda) * (exp(gamma * t)-1) + 2 * gamma
    p = 2*kappa*theta/sigmaSqr;
    enum1= 2*gamma*exp(0.5*(kappa+Lambda+gamma)*t)
    A = pow((enum1/denum),p)
    B = (2*(exp(gamma*t)-1))/denum
    dfact=A*exp(-B*r)
    return dfact

def TermStructureDiscountFactorEstimatedCir(t, r, phi1, phi2, phi3):
    tmp = (phi2*(exp(phi1*t)-1.0)+phi1);
    A = (phi1*exp(phi2*t))/tmp;
    A = pow(A,phi3);
    B = (exp(phi1*t)-1.0)/tmp;
    dfact =  A*exp(-B*r);       
    return dfact

if __name__ == '__main__':
    r = 0.1
    t = 1.0
    d = 0.9
    
    print 'r =', r, ' t =', t, ' given disc.fact', TermStructureDiscountFactorFromYield(r, t)
    print 'd =', d, ' t =', t, ' given yield', TermStructureYieldFromDiscountFactor(d, t)

    d1 = 0.8
    print 'implied forward rate d =', d, 'd1 =', d1, 't =', t, ' forward =', TermStructureForwardRateFromDiscountFactor(d, d1, t)

    times  = (0.1, 0.5, 1, 5, 10)
    yields = (0.1, 0.2, 0.3, 0.4, 0.5)
    print 'testing interpolates'
    print 't=0.1 ', TermStructureYieldLinearlyInterpolated(0.1, times, yields)
    print 't=0.5 ', TermStructureYieldLinearlyInterpolated(0.5, times, yields)
    print 't=1 ', TermStructureYieldLinearlyInterpolated(1.0, times, yields)
    print 't=3 ', TermStructureYieldLinearlyInterpolated(3.0, times, yields)
    print 't=5 ', TermStructureYieldLinearlyInterpolated(5.0, times, yields)
    print 't=10 ', TermStructureYieldLinearlyInterpolated(10, times, yields)

    b = 0.1
    c = 0.1
    d = -0.1
    f = (0.01, 0.01, - 0.01)
    knots = (2, 7, 12)
    t = 1.0
    print 'spline discount factor', TermStructureDiscountFactorCubicSpline(t,b,c,d,f,knots)
    t = 1.0
    print 'bliss r(t=1)=', TermStructureYieldBliss(t,0.1,0.1,0.1,1.0,1.0)
    print 'ns =', TermStructureYieldNelsonSiegel(t,0.1,0.1,0.1,1.0)
    print 'vasicek discount factor (t=1)', TermStructureDiscountFactorVasicek(1.0,0.05,-0.1,0.1,0.1)
    print 'cir discount factor (t=1)', TermStructureDiscountFactorCir(1.0,0.1,0.1,0.1,0.1,0.1)
    print 'estimated cir (t=1)', TermStructureDiscountFactorEstimatedCir(1.0,0.1,0.1,0.1,1.1)
    




